pack-objects: don't traverse objects unnecessarily

This brings back some of the performance lost in optimizing recency order inside pack objects. We were doing extreme amounts of object re-traversal: for the 2.14 million objects in the Linux kernel repository, we were calling add_to_write_order() over 1.03 billion times (a 0.2% hit rate, making 99.8% of of these calls extraneous). Two optimizations take place here- we can start our objects array iteration from a known point where we left off before we started trying to find our tags, and we don't need to do the deep dives required by add_family_to_write_order() if the object has already been marked as filled. These two optimizations bring some pretty spectacular results via `perf stat`: task-clock: 83373 ms --> 43800 ms (50% faster) cycles: 221,633,461,676 --> 116,307,209,986 (47% fewer) instructions: 149,299,179,939 --> 122,998,800,184 (18% fewer) Helped-by: Ramsay Jones (format string fix in "die" message) Signed-off-by: Dan McGee <dpmcgee@gmail.com> Signed-off-by: Junio C Hamano <gitster@pobox.com>

Dan McGee committed Oct 18, 2011 at 00:21 UTC 38d4debb6d180ca53fcb12b8115e81fd4c5262d0
1 file changed +12 -6
builtin/pack-objects.c
+12 -6
@@ -520,7 +520,7 @@ static void add_family_to_write_order(struct object_entry **wo,
520
521 static struct object_entry **compute_write_order(void)
522 {
523 - unsigned int i, wo_end;
523 + unsigned int i, wo_end, last_untagged;
524
525 struct object_entry **wo = xmalloc(nr_objects * sizeof(*wo));
526
@@ -551,7 +551,7 @@ static struct object_entry **compute_write_order(void)
551 for_each_tag_ref(mark_tagged, NULL);
552
553 /*
554 - * Give the commits in the original recency order until
554 + * Give the objects in the original recency order until
555 * we see a tagged tip.
556 */
557 for (i = wo_end = 0; i < nr_objects; i++) {
@@ -559,6 +559,7 @@ static struct object_entry **compute_write_order(void)
559 break;
560 add_to_write_order(wo, &wo_end, &objects[i]);
561 }
562 + last_untagged = i;
563
564 /*
565 * Then fill all the tagged tips.
@@ -571,7 +572,7 @@ static struct object_entry **compute_write_order(void)
572 /*
573 * And then all remaining commits and tags.
574 */
574 - for (i = 0; i < nr_objects; i++) {
575 + for (i = last_untagged; i < nr_objects; i++) {
576 if (objects[i].type != OBJ_COMMIT &&
577 objects[i].type != OBJ_TAG)
578 continue;
@@ -581,7 +582,7 @@ static struct object_entry **compute_write_order(void)
582 /*
583 * And then all the trees.
584 */
584 - for (i = 0; i < nr_objects; i++) {
585 + for (i = last_untagged; i < nr_objects; i++) {
586 if (objects[i].type != OBJ_TREE)
587 continue;
588 add_to_write_order(wo, &wo_end, &objects[i]);
@@ -590,8 +591,13 @@ static struct object_entry **compute_write_order(void)
591 /*
592 * Finally all the rest in really tight order
593 */
593 - for (i = 0; i < nr_objects; i++)
594 - add_family_to_write_order(wo, &wo_end, &objects[i]);
594 + for (i = last_untagged; i < nr_objects; i++) {
595 + if (!objects[i].filled)
596 + add_family_to_write_order(wo, &wo_end, &objects[i]);
597 + }
598 +
599 + if (wo_end != nr_objects)
600 + die("ordered %u objects, expected %"PRIu32, wo_end, nr_objects);
601
602 return wo;
603 }